/*
  Dokan : user-mode file system library for Windows

  Copyright (C) 2008 Hiroki Asakawa info@dokan-dev.net

  http://dokan-dev.net/en

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation; either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with this program. If not, see <http://www.gnu.org/licenses/>.
*/

//-----------------------------------------------------------------------------------------------
#include "dokan.h"

#include <mountdev.h>
#include <mountmgr.h>
#include <ntddvol.h>

//-----------------------------------------------------------------------------------------------
//--- PrintUnknownDeviceIoctlCode
//-----------------------------------------------------------------------------------------------
VOID
PrintUnknownDeviceIoctlCode(
	__in ULONG	IoctlCode)
{
	PCHAR baseCodeStr = "unknown";
	ULONG baseCode = DEVICE_TYPE_FROM_CTL_CODE(IoctlCode);
	ULONG functionCode = (IoctlCode & (~0xffffc003)) >> 2;

	DDBGPRINT("[RCloudFS] PrintUnknownDeviceIoctlCode() : ...Unknown Code=[0x%x]", IoctlCode);

	switch (baseCode) 
	{
		case IOCTL_STORAGE_BASE:
			baseCodeStr = "IOCTL_STORAGE_BASE";
			break;

		case IOCTL_DISK_BASE:
			baseCodeStr = "IOCTL_DISK_BASE";
			break;

		case IOCTL_VOLUME_BASE:
			baseCodeStr = "IOCTL_VOLUME_BASE";
			break;

		case MOUNTDEVCONTROLTYPE:
			baseCodeStr = "MOUNTDEVCONTROLTYPE";
			break;

		case MOUNTMGRCONTROLTYPE:
			baseCodeStr = "MOUNTMGRCONTROLTYPE";
			break;
	}

	DDBGPRINT("[RCloudFS] PrintUnknownDeviceIoctlCode() : ...BaseCode=[0x%x(%s)], FunctionCode=[0x%x(%d)]", baseCode, baseCodeStr, functionCode, functionCode);
}

//-----------------------------------------------------------------------------------------------
//--- GlobalDeviceControl
//-----------------------------------------------------------------------------------------------
NTSTATUS
GlobalDeviceControl(
	__in PDEVICE_OBJECT DeviceObject,
	__in PIRP Irp)
{
	PIO_STACK_LOCATION	irpSp;
	NTSTATUS			status = STATUS_NOT_IMPLEMENTED;
	
	DDBGPRINT("[RCloudFS] DokanGlobalDeviceControl() : ...Start");
	irpSp = IoGetCurrentIrpStackLocation(Irp);

	switch (irpSp->Parameters.DeviceIoControl.IoControlCode) 
	{
		case IOCTL_EVENT_START:
			DDBGPRINT("[RCloudFS] DokanGlobalDeviceControl() : ...IOCTL_EVENT_START");
			status = DokanEventStart(DeviceObject, Irp);
			break;

		case IOCTL_SERVICE_WAIT:
			status = DokanRegisterPendingIrpForService(DeviceObject, Irp);
			DDBGPRINT("[RCloudFS] DokanGlobalDeviceControl() : ...IOCTL_SERVICE_WAIT");
			break;

		case IOCTL_SET_DEBUG_MODE:
			DDBGPRINT("[RCloudFS] DokanGlobalDeviceControl() : ...IOCTL_SET_DEBUG_MODE");
			{
				if (irpSp->Parameters.DeviceIoControl.InputBufferLength >= sizeof(ULONG))
				{
					g_Debug = *(ULONG*)Irp->AssociatedIrp.SystemBuffer;
					status = STATUS_SUCCESS;
				}

				DDBGPRINT("[RCloudFS] DokanGlobalDeviceControl() : ...IOCTL_SET_DEBUG_MODE=[%d]", g_Debug);
			}
			break;

		case IOCTL_TEST:
			DDBGPRINT("[RCloudFS] DokanGlobalDeviceControl() : ...IOCTL_TEST");
			if (irpSp->Parameters.DeviceIoControl.OutputBufferLength >= sizeof(ULONG)) 
			{
				*(ULONG*)Irp->AssociatedIrp.SystemBuffer = DOKAN_DRIVER_VERSION;
				Irp->IoStatus.Information = sizeof(ULONG);
				status = STATUS_SUCCESS;
				break;
			}

		default: 
			PrintUnknownDeviceIoctlCode(irpSp->Parameters.DeviceIoControl.IoControlCode);
			status = STATUS_INVALID_PARAMETER;
			break;
	}

	DDBGPRINT("[RCloudFS] DokanGlobalDeviceControl() : ...");
	return status;
}

//-----------------------------------------------------------------------------------------------
//--- DiskDeviceControl
//-----------------------------------------------------------------------------------------------
NTSTATUS
DiskDeviceControl(
	__in PDEVICE_OBJECT DeviceObject,
	__in PIRP Irp)
{
	PIO_STACK_LOCATION	irpSp;
	PDokanDCB			dcb;
	NTSTATUS			status = STATUS_NOT_IMPLEMENTED;
	ULONG				outputLength = 0;
	
	DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...Start");
	irpSp = IoGetCurrentIrpStackLocation(Irp);
	dcb = DeviceObject->DeviceExtension;
	outputLength = irpSp->Parameters.DeviceIoControl.OutputBufferLength;

	switch (irpSp->Parameters.DeviceIoControl.IoControlCode) 
	{
		case IOCTL_DISK_GET_DRIVE_GEOMETRY:
			{
				PDISK_GEOMETRY	diskGeometry;
				ULONG		    length;

				DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_DISK_GET_DRIVE_GEOMETRY");
				if (outputLength < sizeof(DISK_GEOMETRY))
				{
					status = STATUS_BUFFER_TOO_SMALL;
					Irp->IoStatus.Information = 0;
					break;
				}

				diskGeometry = (PDISK_GEOMETRY) Irp->AssociatedIrp.SystemBuffer;
				ASSERT(diskGeometry != NULL);
				length = 1024 * 1024 * 1024;
				diskGeometry->Cylinders.QuadPart = length / DOKAN_SECTOR_SIZE / 32 / 2;
				diskGeometry->MediaType = FixedMedia;
				diskGeometry->TracksPerCylinder = 2;
				diskGeometry->SectorsPerTrack = 32;
				diskGeometry->BytesPerSector = DOKAN_SECTOR_SIZE;

				status = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(DISK_GEOMETRY);
			}

			break;

		case IOCTL_DISK_GET_LENGTH_INFO:
			{
				PGET_LENGTH_INFORMATION getLengthInfo;

				DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_DISK_GET_LENGTH_INFO");
	            
				if (outputLength < sizeof(GET_LENGTH_INFORMATION)) 
				{
					status = STATUS_BUFFER_TOO_SMALL;
					Irp->IoStatus.Information = 0;
					break;
				}

				getLengthInfo = (PGET_LENGTH_INFORMATION) Irp->AssociatedIrp.SystemBuffer;
				ASSERT(getLengthInfo != NULL);

				getLengthInfo->Length.QuadPart = 1024 * 1024 * 500;
				status = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(GET_LENGTH_INFORMATION);
			}

			break;

		case IOCTL_DISK_GET_PARTITION_INFO:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_DISK_GET_PARTITION_INFO");
			break;

		case IOCTL_DISK_GET_PARTITION_INFO_EX:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_DISK_GET_PARTITION_INFO_EX");
			break;

		case IOCTL_DISK_IS_WRITABLE:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_DISK_IS_WRITABLE");
			status = STATUS_SUCCESS;
			break;

		case IOCTL_DISK_MEDIA_REMOVAL:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_DISK_MEDIA_REMOVAL");
			status = STATUS_SUCCESS;
			break;

		case IOCTL_STORAGE_MEDIA_REMOVAL:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_STORAGE_MEDIA_REMOVAL");
			status = STATUS_SUCCESS;
			break;

		case IOCTL_DISK_SET_PARTITION_INFO:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_DISK_SET_PARTITION_INFO");
			break;

		case IOCTL_DISK_VERIFY:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_DISK_VERIFY");
			break;

		case IOCTL_STORAGE_GET_HOTPLUG_INFO:
			{
				PSTORAGE_HOTPLUG_INFO hotplugInfo;
				DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_STORAGE_GET_HOTPLUG_INFO");
				if (outputLength < sizeof(STORAGE_HOTPLUG_INFO)) 
				{
					status = STATUS_BUFFER_TOO_SMALL;
					Irp->IoStatus.Information = 0;
					break;
				}

				hotplugInfo = Irp->AssociatedIrp.SystemBuffer;
				hotplugInfo->Size =  sizeof(STORAGE_HOTPLUG_INFO);
				hotplugInfo->MediaRemovable = 1;
				hotplugInfo->MediaHotplug = 1;
				hotplugInfo->DeviceHotplug = 1;
				hotplugInfo->WriteCacheEnableOverride = 0;
				status = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(STORAGE_HOTPLUG_INFO);
			}
			break;

		case IOCTL_VOLUME_GET_GPT_ATTRIBUTES:
			{
				DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_VOLUME_GET_GPT_ATTRIBUTES");
				status = STATUS_SUCCESS;
			}
			break;
		case IOCTL_DISK_CHECK_VERIFY:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_DISK_CHECK_VERIFY");
			status = STATUS_SUCCESS;
			break;

		case IOCTL_STORAGE_CHECK_VERIFY:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_STORAGE_CHECK_VERIFY");
			status = STATUS_SUCCESS;
			break;

		case IOCTL_STORAGE_CHECK_VERIFY2:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_STORAGE_CHECK_VERIFY2");
			status = STATUS_SUCCESS;
			break;

		case IOCTL_MOUNTDEV_QUERY_DEVICE_NAME:
			{
				PMOUNTDEV_NAME	mountdevName;
				ULONG			bufferLength = irpSp->Parameters.DeviceIoControl.OutputBufferLength;
				PUNICODE_STRING	deviceName =  dcb->DiskDeviceName;
				
				DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_MOUNTDEV_QUERY_DEVICE_NAME");

				if (bufferLength < sizeof(MOUNTDEV_NAME)) 
				{
					status = STATUS_BUFFER_TOO_SMALL;
					Irp->IoStatus.Information = sizeof(MOUNTDEV_NAME);
					break;
				}

				if (!dcb->Mounted)
				{
					status = STATUS_INVALID_PARAMETER;
					break;
				}

				mountdevName = (PMOUNTDEV_NAME) Irp->AssociatedIrp.SystemBuffer;
				ASSERT(mountdevName != NULL);
				/* NOTE: When Windows API GetVolumeNameForVolumeMountPoint is called, this IO control is called.
				   Even if status = STATUS_SUCCESS, GetVolumeNameForVolumeMountPoint returns error.
				   Something is wrong..
				*/
				mountdevName->NameLength = deviceName->Length;

				if (sizeof(USHORT) + mountdevName->NameLength < bufferLength) 
				{
					RtlCopyMemory((PCHAR) mountdevName->Name, deviceName->Buffer, mountdevName->NameLength);
					Irp->IoStatus.Information = sizeof(USHORT) + mountdevName->NameLength;
					status = STATUS_SUCCESS;
					DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...DeviceName=[%wZ]", deviceName);
				} 
				else 
				{
					Irp->IoStatus.Information = sizeof(MOUNTDEV_NAME);
					status = STATUS_BUFFER_OVERFLOW;
				}
			}

			break;

		case IOCTL_MOUNTDEV_QUERY_UNIQUE_ID:	
			{
				PMOUNTDEV_UNIQUE_ID uniqueId;
				ULONG				bufferLength = irpSp->Parameters.DeviceIoControl.OutputBufferLength;

				DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_MOUNTDEV_QUERY_UNIQUE_ID");
				if (bufferLength < sizeof(MOUNTDEV_UNIQUE_ID)) 
				{
					status = STATUS_BUFFER_TOO_SMALL;
					Irp->IoStatus.Information = sizeof(MOUNTDEV_UNIQUE_ID);
					break;
				}

				uniqueId = (PMOUNTDEV_UNIQUE_ID)Irp->AssociatedIrp.SystemBuffer;
				ASSERT(uniqueId != NULL);

				uniqueId->UniqueIdLength = dcb->SymbolicLinkName->Length;

				if (sizeof(USHORT) + uniqueId->UniqueIdLength < bufferLength) 
				{
					RtlCopyMemory((PCHAR)uniqueId->UniqueId, dcb->SymbolicLinkName->Buffer,	uniqueId->UniqueIdLength);
					Irp->IoStatus.Information = FIELD_OFFSET(MOUNTDEV_UNIQUE_ID, UniqueId[0]) +	uniqueId->UniqueIdLength;
					status = STATUS_SUCCESS;
					DDBGPRINT("[RCloudFS] DiskDeviceControl() : FIELD_OFFSET()...UniqueName=[%ws]", uniqueId->UniqueId);
					break;
				} 
				else 
				{
					Irp->IoStatus.Information = sizeof(MOUNTDEV_UNIQUE_ID);
					status = STATUS_BUFFER_OVERFLOW;
				}
			}
			
			break;

		case IOCTL_MOUNTDEV_QUERY_SUGGESTED_LINK_NAME:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_MOUNTDEV_QUERY_SUGGESTED_LINK_NAME");
			break;

		case IOCTL_MOUNTDEV_LINK_CREATED:
			{
				PMOUNTDEV_NAME	mountdevName = Irp->AssociatedIrp.SystemBuffer;
				DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_MOUNTDEV_LINK_CREATED");
				DDBGPRINT("[RCloudFS] DiskDeviceControl() : Name=[%ws]", mountdevName->Name); 
				status = STATUS_SUCCESS;
			}
			break;

		case IOCTL_MOUNTDEV_LINK_DELETED:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : ...IOCTL_MOUNTDEV_LINK_DELETED");
			status = STATUS_SUCCESS;
			break;

		//case IOCTL_MOUNTDEV_UNIQUE_ID_CHANGE_NOTIFY:
		//	DDBGPRINT("[RCloudFS]    IOCTL_MOUNTDEV_UNIQUE_ID_CHANGE_NOTIFY");
		//	break;

		case IOCTL_MOUNTDEV_QUERY_STABLE_GUID:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_MOUNTDEV_QUERY_STABLE_GUID");
			break;

		case IOCTL_VOLUME_ONLINE:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_VOLUME_ONLINE");
			status = STATUS_SUCCESS;
			break;

		case IOCTL_VOLUME_OFFLINE:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_VOLUME_OFFLINE");
			status = STATUS_SUCCESS;
			break;

		case IOCTL_VOLUME_READ_PLEX:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_VOLUME_READ_PLEX");
			break;
		case IOCTL_VOLUME_PHYSICAL_TO_LOGICAL:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_VOLUME_PHYSICAL_TO_LOGICAL");
			break;

		case IOCTL_VOLUME_IS_CLUSTERED:
			DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_VOLUME_IS_CLUSTERED");
			break;

		case IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS:
			{
				PVOLUME_DISK_EXTENTS	volume;
				ULONG	bufferLength = irpSp->Parameters.DeviceIoControl.OutputBufferLength;

				DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS");
				if (bufferLength < sizeof(VOLUME_DISK_EXTENTS)) 
				{
					status =  STATUS_INVALID_PARAMETER;
					Irp->IoStatus.Information = 0;
					break;
				}

				volume = Irp->AssociatedIrp.SystemBuffer;
				RtlZeroMemory(volume, sizeof(VOLUME_DISK_EXTENTS));
				volume->NumberOfDiskExtents = 1;
				Irp->IoStatus.Information = sizeof(VOLUME_DISK_EXTENTS);
				status = STATUS_SUCCESS;
			}

			break;

		case IOCTL_STORAGE_EJECT_MEDIA:
			{
				DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_STORAGE_EJECT_MEDIA");
				DokanUnmount(dcb);				
				status = STATUS_SUCCESS;
			}

			break;

		case IOCTL_REDIR_QUERY_PATH:
			{
				DDBGPRINT("[RCloudFS] DiskDeviceControl() : IOCTL_REDIR_QUERY_PATH");
			}

			break;

		default:
			PrintUnknownDeviceIoctlCode(irpSp->Parameters.DeviceIoControl.IoControlCode);
			status = STATUS_NOT_IMPLEMENTED;
			break;
	}

	DDBGPRINT("[RCloudFS] DokanDiskDeviceControl() : ...End");

	return status;
}

//-----------------------------------------------------------------------------------------------
//--- DokanDispatchDeviceControl
//-----------------------------------------------------------------------------------------------
NTSTATUS
DokanDispatchDeviceControl(
	__in PDEVICE_OBJECT DeviceObject,
	__in PIRP Irp)

/*++

Routine Description:

	This device control dispatcher handles IOCTLs.

Arguments:

	DeviceObject - Context for the activity.
	Irp 		 - The device control argument block.

Return Value:

	NTSTATUS

--*/

{
	PDokanVCB			vcb;
	PDokanDCB			dcb;
	PIO_STACK_LOCATION	irpSp;
	NTSTATUS			status = STATUS_NOT_IMPLEMENTED;
	ULONG				controlCode;
	// {DCA0E0A5-D2CA-4f0f-8416-A6414657A77A}
	GUID dokanGUID = { 0xdca0e0a5, 0xd2ca, 0x4f0f, { 0x84, 0x16, 0xa6, 0x41, 0x46, 0x57, 0xa7, 0x7a } };

	__try 
	{
		FsRtlEnterFileSystem();

		Irp->IoStatus.Information = 0;

		irpSp = IoGetCurrentIrpStackLocation(Irp);

		controlCode = irpSp->Parameters.DeviceIoControl.IoControlCode;
	
		if (controlCode != IOCTL_EVENT_WAIT && controlCode != IOCTL_EVENT_INFO && controlCode != IOCTL_KEEPALIVE) 
		{
			DDBGPRINT("[RCloudFS] DokanDispatchDeviceControl() : ...Start");
			DDBGPRINT("[RCloudFS] DokanDispatchDeviceControl() : IoGetRequestorProcessId()...ProcessId=[%lu]", IoGetRequestorProcessId(Irp));
		}

		vcb = DeviceObject->DeviceExtension;
		if (GetIdentifierType(vcb) == DGL) 
		{
			status = GlobalDeviceControl(DeviceObject, Irp);
			
			__leave;
		} 
		else if (GetIdentifierType(vcb) == DCB) 
		{
			status = DiskDeviceControl(DeviceObject, Irp);
			
			__leave;
		} 
		else if (GetIdentifierType(vcb) != VCB) 
		{
			status = STATUS_INVALID_PARAMETER;
			
			__leave;
		}
		dcb = vcb->Dcb;

		switch (irpSp->Parameters.DeviceIoControl.IoControlCode) 
		{
			case IOCTL_EVENT_WAIT:
				DDBGPRINT("[RCloudFS] DokanDispatchDeviceControl() : ...IOCTL_EVENT_WAIT");
				status = DokanRegisterPendingIrpForEvent(DeviceObject, Irp);
				break;

			case IOCTL_EVENT_INFO:
				DDBGPRINT("[RCloudFS] DokanDispatchDeviceControl() : ...IOCTL_EVENT_INFO");
				status = DokanCompleteIrp(DeviceObject, Irp);
				break;

			case IOCTL_EVENT_RELEASE:
				DDBGPRINT("[RCloudFS] DokanDispatchDeviceControl() : ...IOCTL_EVENT_RELEASE");
				status = DokanEventRelease(DeviceObject);
				break;

			case IOCTL_EVENT_WRITE:
				DDBGPRINT("[RCloudFS] DokanDispatchDeviceControl() : ...IOCTL_EVENT_WRITE");
				status = DokanEventWrite(DeviceObject, Irp);
				break;

			case IOCTL_KEEPALIVE:
				if (dcb->Mounted) 
				{
					ExAcquireResourceExclusiveLite(&dcb->Resource, TRUE);
					DokanUpdateTimeout(&dcb->TickCount, DOKAN_KEEPALIVE_TIMEOUT);
					ExReleaseResourceLite(&dcb->Resource);
					status = STATUS_SUCCESS;
				} 
				else 
				{
					DDBGPRINT("[RCloudFS]  device is not mounted");
					status = STATUS_INSUFFICIENT_RESOURCES;
				}
				break;

			case IOCTL_RESET_TIMEOUT:
				DDBGPRINT("[RCloudFS] DokanDispatchDeviceControl() : ...IOCTL_RESET_TIMEOUT");
				status = DokanResetPendingIrpTimeout(DeviceObject, Irp);
				break;
			
			//--- 20121204
			case IOCTL_GET_ACCESS_TOKEN:
				DDBGPRINT("[RCloudFS] DokanDispatchDeviceControl() : ...IOCTL_GET_ACCESS_TOKEN");
				status = DokanGetAccessToken(DeviceObject, Irp);
				break;

			default:
				{
					PrintUnknownDeviceIoctlCode(irpSp->Parameters.DeviceIoControl.IoControlCode);
					status = STATUS_NOT_IMPLEMENTED;
				}

				break;
		} // switch IoControlCode
	
	} 
	__finally 
	{
		if (status != STATUS_PENDING) 
		{
			//
			// complete the Irp
			//
			Irp->IoStatus.Status = status;
			IoCompleteRequest(Irp, IO_NO_INCREMENT);
		}

		if (controlCode != IOCTL_EVENT_WAIT && controlCode != IOCTL_EVENT_INFO && controlCode != IOCTL_KEEPALIVE) 
		{
			DokanPrintNTStatus(status);
			DDBGPRINT("[RCloudFS] DokanDispatchDeviceControl() : ...End");
		}

		FsRtlExitFileSystem();
	}

	return status;
}

//-----------------------------------------------------------------------------------------------